
from django.contrib.auth.models import User, Group
from django.urls import reverse
from django.test import RequestFactory, TestCase
from rest_framework.test import APIClient
from uhepp_vault.models import Collection, Plot

from uhepp_vault.models import PUBLIC_LEVEL, INTERNAL_LEVEL, PRIVATE_LEVEL

class CollectionsEndpointTestCase(TestCase):
    """Test the /collections endpoint of the api"""

    fixtures = ["users.json", "collections.json"]

    def setUp(self):
        self.client = APIClient()
        self.frank = User.objects.get(username="cern-fsauerbu")
        self.karl = User.objects.get(username="cern-karl")
        self.john = User.objects.get(username="cern-john")

    # List endpoint
    def test_collec_list_anon(self):
        """Check that public collections are shown to anonymous users"""
        response = self.client.get(reverse("collection-list"))
        titles = sorted(collection["title"] for collection in response.json())
        self.assertEqual(titles, ["Frank at the zoo",
                                  "John at the zoo",
                                  "Karl at the zoo"])
        
    def test_collec_list_auth(self):
        """Check that internal and own collections are shown to auth users"""
        self.client.force_authenticate(self.frank)
        response = self.client.get(reverse("collection-list"))
        titles = sorted(collection["title"] for collection in response.json())
        self.assertEqual(titles, ["Frank at home", "Frank at the zoo",
                                  "Frank in the office", "John at the zoo",
                                  "John in the office", "Karl at the zoo",
                                  "Karl in the office"])


    # Collection detail
    def test_collec_detail_anon_access_public(self):
        """Check that anonymous users can access public collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank at the zoo")

    def test_collec_detail_anon_access_internal(self):
        """Check that anonymous users cannot access internal collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_anon_access_private(self):
        """Check that anonymous users cannot access private collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)


    def test_collec_detail_auth_access_public(self):
        """Check that authorized users can access public collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL
        )

        self.client.force_authenticate(self.karl)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank at the zoo")

    def test_collec_detail_auth_access_internal(self):
        """Check that authorized users can access internal collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL
        )

        self.client.force_authenticate(self.karl)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank in the office")

    def test_collec_detail_auth_access_private(self):
        """Check that authorized users cannot access private collections"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL
        )

        self.client.force_authenticate(self.karl)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)


    def test_collec_detail_self_access_public(self):
        """Check that one can access one's own public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        self.client.force_authenticate(self.frank)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank at the zoo")


    def test_collec_detail_self_access_internal(self):
        """Check that one can access one's own internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        self.client.force_authenticate(self.frank)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank in the office")

    def test_collec_detail_self_access_private(self):
        """Check that one can access one's own private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        self.client.force_authenticate(self.frank)

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "Frank at home")

    # Delete
    def test_collec_detail_anon_delete_public(self):
        """Check that anonymous users cannot delete a public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 401)

    def test_collec_detail_anon_delete_internal(self):
        """Check that anonymous users cannot delete a internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_anon_delete_private(self):
        """Check that anonymous users cannot delete a private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_auth_delete_public(self):
        """Check that auth users cannot delete a public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 403)

    def test_collec_detail_auth_delete_internal(self):
        """Check that auth users cannot delete a internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 403)

    def test_collec_detail_auth_delete_private(self):
        """Check that auth users cannot delete a private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_self_delete_public(self):
        """Check that one can delete one's own public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 204)

    def test_collec_detail_auth_delete_internal(self):
        """Check that one can delete one's own internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 204)

    def test_collec_detail_auth_delete_private(self):
        """Check that one can delete one's own private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.delete(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 204)


    # Delete
    def test_collec_detail_anon_update_public(self):
        """Check that anonymous users cannot delete a public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 401)

    def test_collec_detail_anon_update_internal(self):
        """Check that anonymous users cannot delete a internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_anon_update_private(self):
        """Check that anonymous users cannot delete a private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_auth_update_public(self):
        """Check that auth users cannot delete a public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 403)

    def test_collec_detail_auth_update_internal(self):
        """Check that auth users cannot delete a internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 403)

    def test_collec_detail_auth_update_private(self):
        """Check that auth users cannot delete a private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        self.client.force_authenticate(self.karl)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]))

        self.assertEqual(response.status_code, 404)

    def test_collec_detail_self_update_public(self):
        """Check that one can delete one's own public collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(title="Hello")
        )

        self.assertEqual(response.status_code, 200)

    def test_collec_detail_auth_update_internal(self):
        """Check that one can delete one's own internal collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(title="Hello")
        )

        self.assertEqual(response.status_code, 200)

    def test_collec_detail_auth_update_private(self):
        """Check that one can delete one's own private collection"""
        collection = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )

        self.client.force_authenticate(self.frank)
        response = self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(title="Hello")
        )

        self.assertEqual(response.status_code, 200)

    def test_collec_detail_overwrite_user_auth(self):
        """Check that collections cannot be stolen"""
        frank_url = reverse("user-detail", args=[self.frank.pk])

        collection = Collection.objects.get(
            owner=self.karl,
            visibility=PUBLIC_LEVEL,
        )

        # Try to update user
        self.client.force_authenticate(self.frank)
        self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(owner=frank_url)
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]),
        )
        self.assertIn("cern-karl", response.json()["owner"])

    def test_collec_detail_overwrite_user_anon(self):
        """Check that collections cannot be stolen by anon"""
        frank_url = reverse("user-detail", args=[self.frank.pk])

        collection = Collection.objects.get(
            owner=self.karl,
            visibility=PUBLIC_LEVEL,
        )

        # Try to update user
        self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(owner=frank_url)
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]),
        )
        self.assertIn("cern-karl", response.json()["owner"])

    def test_collec_detail_overwrite_user_self(self):
        """Check that collections cannot be handed over"""
        frank_url = reverse("user-detail", args=[self.frank.pk])

        collection = Collection.objects.get(
            owner=self.karl,
            visibility=PUBLIC_LEVEL,
        )

        # Try to update user
        self.client.force_authenticate(self.karl)
        self.client.put(
            reverse("collection-detail", args=[collection.pk]),
            data=dict(owner=frank_url)
        )

        response = self.client.get(
            reverse("collection-detail", args=[collection.pk]),
        )
        self.assertIn("cern-karl", response.json()["owner"])


    # Create
    def test_collec_detail_anon_create(self):
        """Check that anonymous users cannot create a collection"""
        response = self.client.post(
            reverse("collection-list"),
            data=dict(title="Hello")
        )

        self.assertEqual(response.status_code, 401)

    def test_collec_detail_auth_create(self):
        """Check that auth users can create a collection"""
        self.client.force_authenticate(self.frank)
        response = self.client.post(
            reverse("collection-list"),
            data=dict(title="Hello")
        )

        self.assertEqual(response.status_code, 201)
        self.assertIn("cern-fsauerbu", response.json()["owner"])

    def test_collec_detail_anon_create_other(self):
        """Check that anon users cannot create a collection for other users"""
        frank_url = reverse("user-detail", args=[self.frank.pk])

        response = self.client.post(
            reverse("collection-list"),
            data=dict(title="Hello", owner=frank_url)
        )

        self.assertEqual(response.status_code, 401)

    def test_collec_detail_auth_create_other(self):
        """Check that auth users cannot create a collection for other users"""
        frank_url = reverse("user-detail", args=[self.frank.pk])

        self.client.force_authenticate(self.karl)
        response = self.client.post(
            reverse("collection-list"),
            data=dict(title="Hello", owner=frank_url)
        )

        self.assertEqual(response.status_code, 201)
        self.assertIn("cern-karl", response.json()["owner"])

    # Detail plots
    def create_plots(self):
        """Add two plots for all of frank's collections"""
        self.col_public = Collection.objects.get(
            owner=self.frank,
            visibility=PUBLIC_LEVEL,
        )
        self.plot_pub_1 = Plot.objects.create(
            uhepp={},
            collection=self.col_public,
        )
        self.plot_pub_2 = Plot.objects.create(
            uhepp={},
            collection=self.col_public,
        )

        self.col_internal = Collection.objects.get(
            owner=self.frank,
            visibility=INTERNAL_LEVEL,
        )
        self.plot_int_1 = Plot.objects.create(
            uhepp={},
            collection=self.col_internal,
        )
        self.plot_int_2 = Plot.objects.create(
            uhepp={},
            collection=self.col_internal,
        )

        self.col_private = Collection.objects.get(
            owner=self.frank,
            visibility=PRIVATE_LEVEL,
        )
        self.plot_priv_1 = Plot.objects.create(
            uhepp={},
            collection=self.col_private,
        )
        self.plot_priv_2 = Plot.objects.create(
            uhepp={},
            collection=self.col_private,
        )
        

    @staticmethod
    def plot_in_list(plot, list):
        """True if the reverse url of the plot is a in exactly one item"""
        plot_url = reverse("plot-detail", args=[plot.uuid])
        return sum(plot_url in url for url in list) == 1

    def test_collec_detail_anon_public_plots(self):
        """Check that plots of a public collection are listed"""
        self.create_plots()
        response = self.client.get(
            reverse("collection-detail", args=[self.col_public.pk])
        )
        plot_urls = response.json()["plots"]

        self.assertTrue(self.plot_in_list(self.plot_pub_1, plot_urls))
        self.assertTrue(self.plot_in_list(self.plot_pub_2, plot_urls))
        self.assertEqual(len(plot_urls), 2)

    def test_collec_detail_auth_internal_plots(self):
        """Check that plots of a internal collection are listed"""
        self.create_plots()
        self.client.force_authenticate(self.john)
        response = self.client.get(
            reverse("collection-detail", args=[self.col_internal.pk])
        )
        plot_urls = response.json()["plots"]

        self.assertTrue(self.plot_in_list(self.plot_int_1, plot_urls))
        self.assertTrue(self.plot_in_list(self.plot_int_2, plot_urls))
        self.assertEqual(len(plot_urls), 2)

    def test_collec_detail_self_private_plots(self):
        """Check that plots of a private collection are listed"""
        self.create_plots()
        self.client.force_authenticate(self.frank)
        response = self.client.get(
            reverse("collection-detail", args=[self.col_private.pk])
        )

        plot_urls = response.json()["plots"]

        self.assertTrue(self.plot_in_list(self.plot_priv_1, plot_urls))
        self.assertTrue(self.plot_in_list(self.plot_priv_2, plot_urls))
        self.assertEqual(len(plot_urls), 2)

    # Headers
    def test_collec_create_headers(self):
        """Check that create response contains the correct headers"""
        self.client.force_authenticate(self.frank)
        response = self.client.post(
            reverse("collection-list"),
            {'title': "Hello"}
        )

        url = response.json()["url"]
        id = str(response.json()["id"])

        self.assertIn(url, response["location"])
        self.assertIn(id, response["link"])
        self.assertIn("ui", response["link"])

    def test_collec_detail_headers(self):
        """Check that detail response contains the correct headers"""
        collection = Collection.objects.create(title="Hello", owner=self.frank)
    
        self.client.force_authenticate(self.frank)
        url = reverse("collection-detail", args=[collection.id])
        response = self.client.get(url)

        self.assertIn(str(collection.id), response["link"])
        self.assertIn("ui", response["link"])
